import typing as t

from ml_pipeline_engine.artifact_store.store.no_op import NoOpArtifactStore
from ml_pipeline_engine.cache import Cache
from ml_pipeline_engine.events import EventSourceMixin
from ml_pipeline_engine.module_loading import get_instance
from ml_pipeline_engine.node import generate_pipeline_id
from ml_pipeline_engine.types import (
    ArtifactStoreLike,
    CaseResult,
    EventManagerLike,
    ModelName,
    NodeId,
    PipelineChartLike,
    PipelineId,
)


class DAGPipelineContext(EventSourceMixin):
    """
    Контекст выполнения пайплайна ML-модели
    """

    def __init__(
        self,
        chart: PipelineChartLike,
        pipeline_id: PipelineId = None,
        input_kwargs: t.Dict[str, t.Any] = None,
        meta: t.Dict[str, t.Any] = None,
    ):
        self.chart = chart
        self.pipeline_id = pipeline_id if pipeline_id is not None else generate_pipeline_id()
        self.input_kwargs = input_kwargs if input_kwargs is not None else {}
        self.meta = meta if meta is not None else {}
        self.artifact_store: ArtifactStoreLike = get_instance(
            cls=self.chart.artifact_store or NoOpArtifactStore, ctx=self
        )

        self._local_cache = self.get_cache_object()
        self._event_managers = [get_instance(cls) for cls in self.chart.event_managers]
        self._case_results = {}
        self._nodes_in_run = set()
        self._active_recurrence_subgraph = {}

    @classmethod
    def get_cache_object(cls):
        return Cache()

    async def save_node_result(self, node_id: NodeId, data: t.Any) -> None:
        self._local_cache.save(node_id=node_id, data=data)
        await self.artifact_store.save(node_id=node_id, data=data)

    async def load_node_result(self, node_id: NodeId) -> t.Any:
        return self._local_cache.load(node_id=node_id)

    async def add_case_result(self, switch_node_id: NodeId, selection: CaseResult) -> None:
        self._case_results[switch_node_id] = selection

    async def get_case_result(self, switch_node_id: NodeId) -> CaseResult:
        return self._case_results[switch_node_id]

    async def add_node_in_run(self, node_id: NodeId) -> None:
        self._nodes_in_run.add(node_id)

    async def is_node_in_run(self, node_id: NodeId) -> bool:
        return node_id in self._nodes_in_run

    async def exists_node_result(self, node_id: NodeId) -> bool:
        return self._local_cache.exists(node_id)

    async def is_active_recurrence_subgraph(self, source: NodeId, dest: NodeId) -> bool:
        return (source, dest) in self._active_recurrence_subgraph

    async def set_active_recurrence_subgraph(self, source: NodeId, dest: NodeId) -> None:
        self._active_recurrence_subgraph[(source, dest)] = 1

    async def remove_recurrence_subgraph(self, source: NodeId, dest: NodeId) -> None:
        self._active_recurrence_subgraph.pop((source, dest))

    def delete_node_results(self, node_ids: t.Iterable[NodeId]) -> None:
        """
        Удаление всей информации, связанной с узлами
        """

        for node_id in node_ids:
            self._local_cache.remove(node_id)
            self._nodes_in_run.discard(node_id)

    @property
    def model_name(self) -> ModelName:
        return self.chart.model_name

    def _get_event_managers(self) -> t.List[EventManagerLike]:
        return self._event_managers

    def __repr__(self):
        return f'<{self.__class__.__name__} model_name="{self.chart.model_name}" pipeline_id="{self.pipeline_id}">'


def create_context_from_chart(
    chart: PipelineChartLike,
    input_kwargs: t.Dict[str, t.Any],
    pipeline_id: PipelineId = None,
    meta: t.Dict[str, t.Any] = None,
) -> DAGPipelineContext:
    """
    Создать контекст выполнения пайплайна ML-модели
    """

    return DAGPipelineContext(
        pipeline_id=pipeline_id, chart=chart, input_kwargs=input_kwargs, meta=meta,
    )



import typing as t

from ml_pipeline_engine.types import NodeBase


class DataSource(NodeBase):
    """
    Базовый класс для источников данных
    """

    node_type = 'datasource'

    def collect(self, *args, **kwargs) -> t.Any:
        raise NotImplementedError('Method collect() is not implemented')



import logging

logger_manager = logging.getLogger('pipeline_engine.manager')
logger_decorators = logging.getLogger('pipeline_engine.decorators')
logger_parallelism = logging.getLogger('pipeline_engine.parallelism')



import typing as t

from ml_pipeline_engine.types import (
    EventManagerLike,
    NodeId,
    PipelineContextLike,
    PipelineResult,
)


class EventManagerBase:
    async def on_pipeline_start(self, ctx: PipelineContextLike) -> None:
        ...

    async def on_pipeline_complete(self, ctx: PipelineContextLike, result: PipelineResult) -> None:
        ...

    async def on_node_start(self, ctx: PipelineContextLike, node_id: NodeId) -> None:
        ...

    async def on_node_complete(self, ctx: PipelineContextLike, node_id: NodeId, error: t.Optional[Exception]) -> None:
        ...


class EventSourceMixin:
    _get_event_managers: t.Callable[..., t.List[t.Type[EventManagerLike]]]

    async def _emit(self, event_name: str, **kwargs):
        for mgr in self._get_event_managers():
            callback = getattr(mgr, event_name, None)

            if callback:
                await callback(ctx=self, **kwargs)

    async def emit_on_node_start(self, node_id: NodeId):
        await self._emit('on_node_start', node_id=node_id)

    async def emit_on_node_complete(self, node_id: NodeId, error: t.Optional[Exception]):
        await self._emit('on_node_complete', node_id=node_id, error=error)

    async def emit_on_pipeline_start(self):
        await self._emit('on_pipeline_start')

    async def emit_on_pipeline_complete(self, result: PipelineResult):
        await self._emit('on_pipeline_complete', result=result)



from concurrent.futures import ProcessPoolExecutor
from multiprocessing import Manager, get_context
from typing import Optional

from ml_pipeline_engine.logs import logger_parallelism as logger
from ml_pipeline_engine.parallelism.basic import (
    PoolExecutorRegistry as BasePoolExecutorRegistry,
)

__all__ = ('process_pool_registry', )


class PoolExecutorRegistry(BasePoolExecutorRegistry):

    def __init__(self):
        super().__init__()
        self._process_manager: Optional[Manager] = None

    def is_ready(self) -> None:

        if not self._pool_executor or self._pool_executor._shutdown_thread:  # noqa
            raise RuntimeError('Исполнение невозможно без указания пула процессов')

        if not self._process_manager:
            raise RuntimeError('Исполнение невозможно без указания менеджера данных для процессов')

    def register_manager(self, manager: Manager) -> None:
        if self._process_manager:
            logger.info(
                'Регистрация менеджера пула процессов возможна только один раз. '
                'Повторная инициализация будет пропущена',
            )
            return

        logger.info('Зарегистрирован менеджер данных для процессов')
        self._process_manager = manager

    def get_manager(self) -> Manager:
        self.is_ready()
        return self._process_manager

    def shutdown(self) -> None:
        super().shutdown()

        if self._process_manager is not None:
            self._process_manager.shutdown()

    def auto_init(self) -> None:
        self.register_manager(Manager())
        self.register_pool_executor(ProcessPoolExecutor(mp_context=get_context('fork')))


process_pool_registry = PoolExecutorRegistry()



from ml_pipeline_engine.types import NodeBase


class FeatureVectorizerBase(NodeBase):
    """
    Базовый класс для векторизаторов
    """

    node_type = 'vectorizer'

    def vectorize(self, *args, **kwargs):
        raise NotImplementedError('Method vectorize() is not implemented')



import typing as t
from abc import ABCMeta, abstractmethod

from ml_pipeline_engine.artifact_store.enums import DataFormat
from ml_pipeline_engine.types import NodeId, PipelineContextLike


class ArtifactStore(metaclass=ABCMeta):
    def __init__(self, ctx: PipelineContextLike, *args, **kwargs):
        self.ctx = ctx

    @abstractmethod
    async def save(self, node_id: NodeId, data: t.Any) -> None:
        ...

    @abstractmethod
    async def load(self, node_id: NodeId) -> t.Any:
        ...


class SerializedArtifactStore(ArtifactStore, metaclass=ABCMeta):
    def __init__(self, ctx: PipelineContextLike, *args, **kwargs):
        super().__init__(ctx=ctx, *args, **kwargs)

    @abstractmethod
    async def save(self, node_id: NodeId, data: t.Any, fmt: DataFormat = None) -> None:
        ...



import copy
import inspect
import typing as t
from collections import deque

from ml_pipeline_engine.dag import DAG, EdgeField, NodeField
from ml_pipeline_engine.dag.graph import DiGraph
from ml_pipeline_engine.dag.utils import get_connected_subgraph
from ml_pipeline_engine.dag_builders.annotation import errors
from ml_pipeline_engine.dag_builders.annotation.marks import (
    InputGenericMark,
    InputMark,
    InputOneOfMark,
    RecurrentSubGraphMark,
    SwitchCaseMark,
)
from ml_pipeline_engine.node import (
    NodeTag,
    generate_node_id,
    get_callable_run_method,
    get_node_id,
)
from ml_pipeline_engine.types import (
    DAGLike,
    NodeBase,
    NodeId,
    NodeLike,
    RecurrentProtocol,
)

__all__ = [
    'build_dag',
    'build_dag_single',
]

KwargName = str

NodeInputSpec = t.Tuple[KwargName, t.Union[InputMark, SwitchCaseMark]]

NodeResultT = t.TypeVar('NodeResultT')


class AnnotationDAGBuilder:
    def __init__(self):
        self._dag = DiGraph()
        self._node_map: t.Dict[NodeId, NodeLike] = dict()
        self._recurrent_sub_graphs: t.List[t.Tuple[NodeId, NodeId]] = []
        self._synthetic_nodes: t.List[NodeId] = []

    @staticmethod
    def _check_annotations(obj: t.Any) -> None:
        """
        Проверка наличия аннотаций типов у переданного объекта.
        В случае, если есть хотя бы один не типизированный параметр, будет ошибка.
        """

        obj = get_callable_run_method(obj)

        annotations = getattr(obj, '__annotations__', None)
        parameters = [
            (name, bool(parameter.empty))
            for name, parameter in inspect.signature(obj).parameters.items()
            if name not in ('self', 'args', 'kwargs')
        ]

        if not annotations and parameters:
            raise errors.UndefinedAnnotation(f'Невозможно найти аннотации типов. obj={obj}')

        for name, is_empty in parameters:
            if is_empty and name not in annotations:
                raise errors.UndefinedParamAnnotation(f'Не указан тип для параметра name={name}, obj={obj}')

    @staticmethod
    def _check_base_class(node: t.Any) -> None:
        """
        Проверка объекта на наличие корректного базового класса у всех узлов
        """

        if not inspect.isclass(node):
            raise errors.IncorrectTypeClass(f'{node} должен быть классом')

        if NodeBase not in inspect.getmro(node):
            raise errors.IncorrectBaseClass(
                f'У объекта не существует корректного базового класса, пригодного для графа. node={node}',
            )

    def validate_node(self, node: NodeLike) -> None:
        """
        Валидация ноды по разным правилам
        """

        self._check_base_class(node)
        self._check_annotations(node)

    @staticmethod
    def _get_input_marks_map(node: NodeLike) -> t.List[NodeInputSpec]:
        """
        Получение меток зависимостей для входных kwarg-ов узла
        """

        node = get_callable_run_method(node)

        inputs = []
        for name, annotation in node.__annotations__.items():  # noqa

            if isinstance(annotation, InputGenericMark):
                raise errors.NonRedefinedGenericTypeError(
                    f'Для использования узлов общего назначения необходимо их переопределение для целевого графа. '
                    f'param_name={name}, node={node}, ',
                )

            if not isinstance(annotation, (InputMark, SwitchCaseMark, InputOneOfMark, RecurrentSubGraphMark)):
                continue

            inputs.append((name, annotation))

        return inputs

    def _add_node_to_map(self, node: NodeLike) -> None:
        """
        Добавление узла в мэппинг "Имя узла -> Класс/функция узла"
        """

        self._node_map[get_node_id(node)] = node

    def _add_node_pair_to_dag(self, source_node_id: NodeId, dest_node_id: NodeId, **edge_data) -> None:
        """
        Добавить в граф пару узлов, связанных ребром
        """

        self._dag.add_node(source_node_id)
        self._dag.add_node(dest_node_id)

        self._dag.add_edge(source_node_id, dest_node_id, **edge_data)

    def _add_switch_node(self, node_id: NodeId, switch_decide_node_id: NodeId) -> None:
        """
        Добавить в граф узел типа switch
        """

        self._dag.add_node(node_id, **{NodeField.is_switch: True})
        self._dag.add_edge(switch_decide_node_id, node_id, **{EdgeField.is_switch: True})

    def _traverse_breadth_first_to_dag(self, input_node: NodeLike, output_node: NodeLike):  # noqa
        """
        Выполнить обход зависимостей классов/функций узлов, построить граф
        """

        visited = {output_node}
        stack = deque([output_node])

        def _set_visited(node: NodeLike) -> None:
            if node in visited:
                return

            visited.add(node)
            stack.append(node)

        while stack:
            current_node = stack.pop()
            self.validate_node(current_node)

            self._add_node_to_map(current_node)

            input_marks_map = self._get_input_marks_map(current_node)

            if not input_marks_map and input_node != current_node:
                self._add_node_pair_to_dag(get_node_id(input_node), get_node_id(current_node))
                _set_visited(input_node)

            for kwarg_name, input_mark in input_marks_map:

                if isinstance(input_mark, RecurrentSubGraphMark):
                    self._add_node_to_map(input_mark.dest_node)
                    self._dag.add_node(
                        get_node_id(input_mark.dest_node),
                        **{
                            NodeField.start_node: get_node_id(input_mark.start_node),
                            NodeField.max_iterations: input_mark.max_iterations,
                        },
                    )
                    self._add_node_pair_to_dag(
                        get_node_id(input_mark.dest_node),
                        get_node_id(current_node),
                        **{EdgeField.kwarg_name: kwarg_name},
                    )
                    self._recurrent_sub_graphs.append(
                        (
                            get_node_id(input_mark.start_node),
                            get_node_id(input_mark.dest_node),
                        ),
                    )
                    _set_visited(input_mark.dest_node)

                if isinstance(input_mark, InputOneOfMark):
                    first_node_in_pool = input_mark.nodes[0]
                    synthetic_node_id = generate_node_id('input_one_of', get_node_id(first_node_in_pool))
                    self._dag.add_node(synthetic_node_id, **{NodeField.is_first_success: True})
                    self._dag.add_edge(get_node_id(input_node), synthetic_node_id)

                    node_list = input_mark.nodes
                    for idx, node in enumerate(node_list):
                        self._add_node_to_map(node)
                        self._dag.add_node(get_node_id(node), **{NodeField.is_first_success_pool: True})

                        if idx + 1 < len(node_list):
                            self._dag.add_edge(
                                get_node_id(node),
                                get_node_id(node_list[idx + 1]),
                                **{EdgeField.is_first_success: get_node_id(first_node_in_pool)},
                            )
                        else:
                            self._dag.add_edge(
                                get_node_id(node),
                                synthetic_node_id,
                                **{EdgeField.is_first_success: get_node_id(first_node_in_pool)},
                            )

                        _set_visited(node)

                    self._synthetic_nodes.append(synthetic_node_id)
                    self._dag.add_edge(
                        synthetic_node_id, get_node_id(current_node), **{EdgeField.kwarg_name: kwarg_name}
                    )

                if isinstance(input_mark, InputMark):
                    self._add_node_to_map(input_mark.node)
                    self._add_node_pair_to_dag(
                        get_node_id(input_mark.node), get_node_id(current_node), **{EdgeField.kwarg_name: kwarg_name}
                    )
                    _set_visited(input_mark.node)

                if isinstance(input_mark, SwitchCaseMark):

                    switch_node_id = generate_node_id('switch', input_mark.name)

                    self._add_node_to_map(input_mark.switch)
                    self._add_switch_node(switch_node_id, get_node_id(input_mark.switch))
                    _set_visited(input_mark.switch)

                    for case_branch, case_node in input_mark.cases:
                        self._add_node_to_map(case_node)
                        self._dag.add_edge(
                            get_node_id(case_node), switch_node_id, **{EdgeField.case_branch: case_branch}
                        )
                        _set_visited(case_node)

                    self._dag.add_edge(switch_node_id, get_node_id(current_node), **{EdgeField.kwarg_name: kwarg_name})
                    self._synthetic_nodes.append(switch_node_id)

    def _validate_recurrent_node_base_classes(self) -> None:
        """
        Валидация узлов из рекуррентных подграфов
        """

        for source, dest in self._recurrent_sub_graphs:
            for node_id in get_connected_subgraph(self._dag, source, dest):

                if node_id in self._synthetic_nodes:
                    continue

                node = self._node_map[node_id]

                if RecurrentProtocol not in inspect.getmro(node):
                    raise errors.IncorrectRecurrentMixinClass(
                        f'{node} не может быть узлом в рекуррентном подграфе, '
                        f'так как не унаследован от {RecurrentProtocol}',
                    )

    def _validate_recurrent_nodes_params(self) -> None:
        """
        Проверка наличия системного параметра в начале подграфа
        """

        for source, dest in self._recurrent_sub_graphs:
            if source in self._synthetic_nodes or dest in self._synthetic_nodes:
                continue

            method = get_callable_run_method(self._node_map[source])

            if 'additional_data' not in method.__annotations__:  # noqa
                raise errors.IncorrectParamsRecurrentNode(
                    f'В {method} отсутствует системный параметр "additional_data" для получения данных от узла, '
                    'который может перезапустить подграф',
                )

            dest_node = self._node_map[dest]
            if not dest_node.use_default:
                raise errors.IncorrectParamsRecurrentNode(
                    f'Для участия в рекуррентном подграфе {dest_node} должен устанавливать параметр use_default=True. '
                    'Дополнительно должен быть переопределен метод get_default()'
                )

    def _validate_graph(self) -> None:
        """
        Метод для валидации построенного графа
        """

        self._validate_recurrent_node_base_classes()
        self._validate_recurrent_nodes_params()

    def _is_executor_needed(self) -> t.Tuple[bool, bool]:
        """
        Проверяем надобность пулов для узлов
        """

        is_thread_pool_needed = False
        is_process_pool_needed = False

        for node in self._node_map.values():

            if inspect.iscoroutinefunction(get_callable_run_method(node)):
                continue

            if NodeTag.process in node.tags:
                is_process_pool_needed = True
            else:
                is_thread_pool_needed = True

        return is_process_pool_needed, is_thread_pool_needed

    def build(self, input_node: NodeLike, output_node: NodeLike = None) -> DAGLike:
        """
        Построить граф путем сборки зависимостей по аннотациям типа (меткам входов)
        """

        self._add_node_to_map(input_node)

        if output_node is None:
            output_node = input_node
            self._dag.add_node(get_node_id(input_node))
        else:
            self._traverse_breadth_first_to_dag(input_node, output_node)

        self._validate_graph()

        is_process_pool_needed, is_thread_pool_needed = self._is_executor_needed()

        return DAG(
            graph=self._dag.copy(),
            input_node=get_node_id(input_node),
            output_node=get_node_id(output_node),
            node_map=copy.deepcopy(self._node_map),
            is_process_pool_needed=is_process_pool_needed,
            is_thread_pool_needed=is_thread_pool_needed,
        )


def build_dag(
    input_node: NodeLike[t.Any],
    output_node: NodeLike[NodeResultT],
) -> DAGLike[NodeResultT]:
    """
    Построить граф путем сборки зависимостей по аннотациям типа (меткам входов)

    Args:
        input_node: Входной узел
        output_node: Выходной узел

    Returns:
        Граф
    """

    return (
        AnnotationDAGBuilder()
        .build(input_node=input_node, output_node=output_node)
    )


def build_dag_single(node: NodeLike[NodeResultT]) -> DAGLike[NodeResultT]:
    """
    Построить граф из одного узла

    Args:
        node: Узел

    Returns:
        Граф
    """
    return AnnotationDAGBuilder().build(input_node=node, output_node=None)



import asyncio
import typing as t
from collections import deque
from contextlib import suppress
from dataclasses import dataclass

import networkx as nx

from ml_pipeline_engine.dag.enums import EdgeField, NodeField
from ml_pipeline_engine.dag.graph import DiGraph
from ml_pipeline_engine.dag.retrying import DagRetryPolicy
from ml_pipeline_engine.dag.utils import get_connected_subgraph
from ml_pipeline_engine.exceptions import NodeErrorType
from ml_pipeline_engine.node import run_node, run_node_default
from ml_pipeline_engine.types import (
    CaseResult,
    DAGLike,
    DAGRunManagerLike,
    NodeId,
    NodeResultT,
    PipelineContextLike,
    Recurrent,
    DAGRunLockManagerLike,
)

from ml_pipeline_engine.logs import logger_manager as logger


class DAGConcurrentManagerLock(DAGRunLockManagerLike):

    def __init__(self, node_ids: t.Iterable[NodeId]):
        self.lock_store = {node_id: asyncio.Event() for node_id in node_ids}

    def get_lock(self, node_id: NodeId) -> asyncio.Event:
        self.lock_store.setdefault(node_id, asyncio.Event())
        return self.lock_store[node_id]


@dataclass
class DAGRunConcurrentManager(DAGRunManagerLike):
    """
    Дефолтный менеджер запуска графов.
    Производит конкурентное исполнение узлов.
    """

    lock_manager: DAGRunLockManagerLike
    dag: DAGLike

    async def run(self, ctx: PipelineContextLike) -> NodeResultT:
        return await self._run_dag(ctx, self._get_reduced_dag(self.dag.input_node, self.dag.output_node))

    async def _get_node_kwargs(self, ctx: PipelineContextLike, node_id: NodeId) -> t.Dict[str, t.Any]:
        """
        Получить входные kwarg-и для узла графа перед его запуском
        """

        kwargs = {}

        if node_id != self.dag.input_node:
            for pred_node_id in self.dag.graph.predecessors(node_id):
                kwarg_name = self.dag.graph.edges[(pred_node_id, node_id)].get(EdgeField.kwarg_name)

                if kwarg_name is None:
                    continue

                if self.dag.graph.nodes[pred_node_id].get(NodeField.is_switch):
                    kwargs[kwarg_name] = await ctx.load_node_result(
                        (await ctx.get_case_result(pred_node_id)).node_id
                    )
                    continue

                kwargs[kwarg_name] = await ctx.load_node_result(pred_node_id)

        else:
            kwargs = ctx.input_kwargs

        additional_data = self.dag.graph.nodes[node_id].get(NodeField.additional_data)

        if additional_data:
            kwargs[NodeField.additional_data] = additional_data

        return kwargs

    @staticmethod
    def _is_switch(dag: DiGraph, node_id: NodeId) -> bool:
        """
        Является ли узел switch-узлом
        """

        try:
            return dag.nodes[node_id].get(NodeField.is_switch) is True
        except KeyError:
            return False

    @staticmethod
    def _is_node_in_oneof(dag: DiGraph, node_id: NodeId) -> bool:
        """
        Является ли узел узлом InputOneOf пула
        """
        return bool(dag.nodes[node_id].get(NodeField.is_first_success_pool))

    @staticmethod
    def _is_head_of_oneof(dag: DiGraph, node_id: NodeId) -> bool:
        """
        Является ли узел родителем InputOneOf пула
        """
        return bool(dag.nodes[node_id].get(NodeField.is_first_success))

    def _get_reduced_dag(self, source: NodeId, dest: NodeId) -> DiGraph:
        """
        Получить связный подграф между двумя заданными узлами графа с удаленными ребрами условных операторов
        """

        def _filter(u, v) -> bool:
            """
            Удаляет ребра с меткой EdgeField.case_branch из subgraph_view

            Args:
                u - Node
                v - Node Edge
            """
            return not self.dag.graph.edges[u, v].get(EdgeField.case_branch)

        def _filter_node(u) -> bool:
            """
            Удаляет ноды с меткой NodeField.is_first_success_pool из subgraph_view

            Args:
                u -  Node
            """
            return not self.dag.graph.nodes[u].get(NodeField.is_first_success_pool)

        return get_connected_subgraph(
            nx.subgraph_view(self.dag.graph, filter_edge=_filter, filter_node=_filter_node), source, dest
        )

    def _get_reduced_dag_input_one_of(self, source: NodeId, dest: NodeId) -> DiGraph:
        """
        Получить связный подграф между двумя заданными узлами графа InputOneOf
        """

        return get_connected_subgraph(nx.subgraph_view(self.dag.graph), source, dest)

    async def _add_case_result(self, ctx: PipelineContextLike, switch_node_id: NodeId) -> None:
        """
        Записать выбор, сделанный оператором выбора Switch-Case, в контекст
        """
        selected_branch_label = None
        branch_nodes = {}
        for pred_id in self.dag.graph.predecessors(switch_node_id):
            edge = self.dag.graph.edges[(pred_id, switch_node_id)]

            if edge.get(EdgeField.is_switch):
                selected_branch_label = await ctx.load_node_result(pred_id)
                continue

            branch_nodes[edge.get(EdgeField.case_branch)] = pred_id

        await ctx.add_case_result(
            switch_node_id,
            CaseResult(label=selected_branch_label, node_id=branch_nodes[selected_branch_label]),
        )

    async def _run_node(
        self,
        ctx: PipelineContextLike,
        node_id: NodeId,
        dag: DiGraph,
        is_node_from_success_pool: bool,
        force_default: bool = False,
    ) -> t.Union[NodeResultT, t.Any]:
        """
        Выполнить узел графа
        """

        # Due to concurrent nature, we have to check if the node is running concurrently in another subgraph.
        # In general, we have two checks: before execution and here. This check prevents a double run.
        # when the deep-nested node in the subgraph has been executed.
        if await ctx.is_node_in_run(node_id):
            logger.debug('Node %s has been executed. Stop new execution', node_id)

            event_lock = self.lock_manager.get_lock(node_id)
            await event_lock.wait()

            return await ctx.load_node_result(node_id)

        await ctx.add_node_in_run(node_id)
        await ctx.emit_on_node_start(node_id=node_id)

        try:
            logger.info('Начало исполнения ноды, node_id=%s', node_id)
            input_kwargs = await self._get_node_kwargs(ctx, node_id)

            # Если в узел попадают параметры, которые являются ошибками, то исполнять этот узел нельзя.
            # Как следствие, нам нужно завершить исполнение текущей ноды
            for key_value in input_kwargs.values():
                if isinstance(key_value, Exception):
                    raise key_value

            result = await self.execute(
                ctx=ctx,
                node_id=node_id,
                force_default=force_default,
                **input_kwargs,
            )

            if isinstance(result, Exception):
                await ctx.emit_on_node_complete(node_id=node_id, error=result)
            else:
                await ctx.emit_on_node_complete(node_id=node_id, error=None)

            logger.info('Завершение исполнения ноды, node_id=%s', node_id)

            if not dag.is_recurrent and await ctx.is_node_in_run(node_id):
                logger.debug('Node %s has been executed. Stop new execution', node_id)

            return result

        except Exception as ex:
            await ctx.emit_on_node_complete(node_id=node_id, error=ex)
            logger.error('Ошибка исполнения ноды node_id=%s, err=%s', node_id, ex)

            if is_node_from_success_pool:
                return NodeErrorType.succession_node_error

            raise ex

    async def execute(
        self,
        ctx: PipelineContextLike,
        node_id: NodeId,
        force_default: bool,
        **kwargs,
    ) -> t.Union[NodeResultT, t.Any]:
        """
        Запуск ноды с дополнительной обработкой

        Args:
            node_id: Идентификатор ноды
            ctx: Контекст исполнения узла
            force_default: Отдать дефолтное значение
            **kwargs: Ключевые аргументы целевой функции
        """

        node = self.dag.node_map[node_id]

        retry_policy = DagRetryPolicy(node=node)

        n_attempts = 1
        while True:
            try:
                logger.debug('Начало исполнения узла node_id=%s', node_id)
                if force_default:
                    return run_node_default(node, **kwargs)

                return await run_node(**kwargs, node=node)

            except retry_policy.exceptions as error:
                logger.debug(
                    'Node %s will be restarted in %s seconds...',
                    node_id,
                    retry_policy.delay,
                    exc_info=error,
                )

                if n_attempts == retry_policy.attempts:

                    if retry_policy.use_default:
                        return run_node_default(node, **kwargs)

                    raise error

                await ctx.emit_on_node_complete(node_id=node_id, error=error)

                n_attempts += 1
                await asyncio.sleep(retry_policy.delay)

            except Exception:
                if retry_policy.use_default:
                    return run_node_default(node, **kwargs)
                raise

    @staticmethod
    async def _get_call_order(ctx: PipelineContextLike, dag: DiGraph) -> t.List[NodeId]:
        """
        Рассчитываем план выполнения узлов графа.
        Учитываем, что если подграф находится в активной рекурсии, то исполнять нужно все узлы
        """

        return [
            node_id for node_id in nx.topological_sort(dag)
            if (not await ctx.is_node_in_run(node_id) if not dag.is_recurrent else True)
        ]

    def _get_node_dependencies(self, dag: DiGraph, node_id: NodeId) -> t.Set[NodeId]:
        """
        Метод возвращает узлы без которых оператор switch-case, oneof не запустится в указанном dag.
        """

        node_predecessors = set(self.dag.graph.predecessors(node_id))
        current_dag = set(nx.topological_sort(dag))

        return current_dag.intersection(node_predecessors)

    async def _get_predecessors(self, dag: DiGraph, ctx: PipelineContextLike, node_id: NodeId) -> t.List[NodeId]:
        """
        Получение всех предшественников для узла
        """
        predecessors = list(
            self._get_node_dependencies(dag, node_id)
            if self._is_switch(dag, node_id) or self._is_head_of_oneof(dag, node_id) or dag.is_recurrent
            else self.dag.graph.predecessors(node_id)
        )

        for idx, node_id in enumerate(predecessors):

            if self._is_switch(dag, node_id):
                with suppress(KeyError):
                    predecessors[idx] = (await ctx.get_case_result(node_id)).node_id

        return predecessors

    async def _is_ready_to_run(self, dag: DiGraph, ctx: PipelineContextLike, node_id: NodeId) -> bool:
        """
        Метод проверяет возможность запуска узла
        """

        # Switch context to collect any predecessors' results to approve the function for running.
        await asyncio.sleep(0)

        if await ctx.is_node_in_run(node_id):
            return False

        for pred_node_id in await self._get_predecessors(dag, ctx, node_id):

            if not await ctx.exists_node_result(pred_node_id):
                return False

        return True

    async def _run_dag(
        self,
        ctx: PipelineContextLike,
        dag: DiGraph,
    ) -> t.Union[NodeResultT, t.Any]:
        """
        Запустить граф / подграф
        """

        logger.debug('Начало запуска DAG, dag=%s', str(dag))

        list_node_ids = await self._get_call_order(ctx, dag)

        if dag.is_recurrent:
            ctx.delete_node_results(list_node_ids)

        if len(list_node_ids) == 0:
            return

        dag_output_node = list_node_ids[-1]
        list_node_ids = deque(list_node_ids)

        awaitable_nodes = []

        while list_node_ids:
            node_id = list_node_ids.popleft()

            # Sometimes we can encounter situations when we have a result of the node-id.
            # In that case, we simply skip the node because:
            #   1. If it's a recurrent subgraph, the result would be deleted for each node in the subgraph
            #      and this if-statement won't be executed.
            #   2. If it's a non-recurrent execution, we just assume that the node's done before in any other graphs.
            if await ctx.is_node_in_run(node_id):
                continue

            elif await self._is_ready_to_run(dag, ctx, node_id):
                logger.debug('Узел готов к обработке node_id=%s', node_id)

                if self._is_switch(dag, node_id):
                    logger.debug('Узел является switch-кейсом node_id=%s', node_id)
                    await self._add_case_result(ctx, node_id)

                    coro_to_run = asyncio.create_task(
                        self._run_dag(
                            ctx=ctx,
                            dag=self._get_reduced_dag(
                                self.dag.input_node,
                                (await ctx.get_case_result(switch_node_id=node_id)).node_id,
                            ),
                        ),
                        name=node_id,
                    )

                elif dag.nodes[node_id].get(NodeField.is_first_success):
                    logger.debug('Узел является InputOneOf node_id=%s', node_id)

                    coro_to_run = asyncio.create_task(
                        self._run_dag(
                            ctx=ctx,
                            dag=self._get_reduced_dag_input_one_of(
                                self.dag.input_node,
                                node_id,
                            ),
                        ),
                        name=node_id,
                    )

                else:
                    logger.debug('Узел может быть обработан, отправляем его в обработку node_id=%s', node_id)
                    coro_to_run = asyncio.create_task(
                        self._run_node(
                            ctx,
                            node_id,
                            # Добавляем флаг, что выполняется нода из InputOneOf пула
                            is_node_from_success_pool=self._is_node_in_oneof(dag, node_id),
                            dag=dag,
                        ),
                        name=node_id,
                    )

                awaitable_nodes.append(coro_to_run)

                # Если мы только запустили последний узел, то результат необходимо получить из соседнего процесса,
                # по этой причине перезапускаем узел для повторной обработки
                if node_id == dag_output_node:
                    list_node_ids.appendleft(node_id)

            else:
                pending_nodes = awaitable_nodes

                while pending_nodes:
                    node_results, pending_nodes = await asyncio.wait(
                        pending_nodes, return_when=asyncio.FIRST_COMPLETED,
                    )

                    for node_result in node_results:
                        node_result_id = node_result.get_name()
                        node_result = await node_result

                        if isinstance(node_result, Recurrent):
                            max_iterations = dag.nodes[node_result_id].get(NodeField.max_iterations)
                            start_from_node_id = dag.nodes[node_result_id].get(NodeField.start_node)

                            if not await ctx.is_active_recurrence_subgraph(start_from_node_id, node_result_id):
                                await ctx.set_active_recurrence_subgraph(start_from_node_id, node_result_id)
                                logger.debug(
                                    'Start the process of the recurrent subgraph for nodes start_node=%s, dest_node=%s',
                                    start_from_node_id,
                                    node_result_id,
                                )

                                recurrent_subgraph = get_connected_subgraph(
                                    self.dag.graph, start_from_node_id, node_result_id, is_recurrent=True,
                                )

                                for current_iter in range(max_iterations):
                                    logger.debug(
                                        'Executing the %s attempt of the recurrent subgraph start_node=%s, dest_node=%s',
                                        current_iter,
                                        start_from_node_id,
                                        node_result_id,
                                    )
                                    self.dag.graph.nodes[start_from_node_id][NodeField.additional_data] = node_result.data

                                    node_result = await self._run_dag(dag=recurrent_subgraph, ctx=ctx)

                                    if current_iter + 1 == max_iterations and isinstance(node_result, Recurrent):

                                        logger.debug(
                                            'Attempts to run a recurrent subgraph have been exceeded. '
                                            'Will be used the default value start_node=%s, dest_node=%s',
                                            start_from_node_id,
                                            node_result_id,
                                        )

                                        ctx.delete_node_results((node_result_id,))
                                        node_result = await self._run_node(
                                            ctx,
                                            node_result_id,
                                            is_node_from_success_pool=self._is_node_in_oneof(dag, node_result_id),
                                            force_default=True,
                                            dag=dag,
                                        )

                                    elif not isinstance(node_result, Recurrent):
                                        break

                                await ctx.remove_recurrence_subgraph(start_from_node_id, node_result_id)

                            else:
                                # В случае, если исполняемый узел рекуррентного подграфа не завершается ожидаемым образом,
                                # то его нужно вернуть обратно к управляющей конструкции
                                logger.debug(
                                    'Finish the process of the recurrent subgraph for nodes '
                                    'start_node=%s, dest_node=%s',
                                    start_from_node_id,
                                    node_result_id,
                                )
                                return node_result

                        # Если нода из списка InputOneOf выполнилась успешно, то проходить остальные не нужно.
                        # При этом, результат сохраняется под выходным узлом
                        if (
                            not isinstance(node_result, NodeErrorType)
                            and self._is_node_in_oneof(dag, node_result_id)
                        ):
                            await ctx.save_node_result(dag_output_node, node_result)
                            # Для искусственной остановки пайплайна помечаем текущий узел как выходной
                            node_result_id = dag_output_node

                        else:
                            # Если узел обычный, то сохраняем результат в обычном режиме под оригинальным node_id
                            await ctx.save_node_result(node_result_id, node_result)

                        self.lock_manager.get_lock(node_result_id).set()

                        if node_result_id == dag_output_node:
                            return await ctx.load_node_result(node_result_id)

                awaitable_nodes.clear()
                list_node_ids.appendleft(node_id)

    def __repr__(self):
        return f'<{self.__class__.__name__} nnodes="{len(self.dag.graph.nodes)}" nedges="{len(self.dag.graph.edges)}">'



import typing as t
from dataclasses import dataclass, field

from ml_pipeline_engine.context import dag as dag_ctx
from ml_pipeline_engine.node import generate_pipeline_id
from ml_pipeline_engine.types import (
    ArtifactStoreLike,
    DAGLike,
    EventManagerLike,
    ModelName,
    NodeLike,
    PipelineId,
    PipelineResult,
)

NodeResultT = t.TypeVar('NodeResultT')

Entrypoint = t.Optional[t.Union[NodeLike[NodeResultT], DAGLike[NodeResultT]]]


@dataclass(frozen=True, repr=False)
class PipelineChartBase:
    """
    Базовый класс определения пайплайна ML-модели
    """

    model_name: ModelName
    entrypoint: Entrypoint
    artifact_store: t.Optional[t.Type[ArtifactStoreLike]] = None
    event_managers: t.List[t.Type[EventManagerLike]] = field(default_factory=list)


@dataclass(frozen=True, repr=False)
class PipelineChart(PipelineChartBase):
    """
    Основная реализация определения пайплайна ML-модели
    """

    async def run(
        self,
        pipeline_id: t.Optional[PipelineId] = None,
        input_kwargs: t.Optional[t.Dict[str, t.Any]] = None,
        meta: t.Optional[t.Dict[str, t.Any]] = None,
    ) -> PipelineResult[NodeResultT]:
        input_kwargs = input_kwargs if input_kwargs is not None else {}
        pipeline_id = pipeline_id if pipeline_id is not None else generate_pipeline_id()

        ctx = dag_ctx.create_context_from_chart(
            chart=self,
            pipeline_id=pipeline_id,
            input_kwargs=input_kwargs,
            meta=meta if meta is not None else {},
        )

        await ctx.emit_on_pipeline_start()

        try:
            result = PipelineResult(
                value=await self.entrypoint.run(ctx),
                pipeline_id=pipeline_id,
                error=None,
            )

            await ctx.emit_on_pipeline_complete(result=result)
            return result

        except Exception as ex:
            result = PipelineResult(pipeline_id=pipeline_id, value=None, error=ex)
            await ctx.emit_on_pipeline_complete(result=result)

            return result


@dataclass(frozen=True, repr=False)
class NullPipelineChart(PipelineChartBase):
    """
    Пустая реализация определения пайплайна ML-модели

    Для обратной совместимости контекста со старым кодом моделей
    """

    entrypoint: Entrypoint = None

    def run(
        self,
        pipeline_id: t.Optional[PipelineId] = None,
        input_kwargs: t.Optional[t.Dict[str, t.Any]] = None,
        meta: t.Dict[str, t.Any] = None,
    ) -> NodeResultT:
        raise NotImplementedError('Unable to run this kind of pipeline')



from dataclasses import dataclass
from typing import Tuple, Type

from ml_pipeline_engine.types import NodeLike, RetryPolicyLike


@dataclass(frozen=True)
class DagRetryPolicy(RetryPolicyLike):
    node: NodeLike

    @property
    def delay(self) -> int:
        return self.node.delay or 0

    @property
    def attempts(self) -> int:
        return self.node.attempts or 1

    @property
    def exceptions(self) -> Tuple[Type[Exception], ...]:
        return self.node.exceptions or (Exception,)

    @property
    def use_default(self) -> bool:
        return self.node.use_default



import functools
import os
import typing as t
import warnings
from enum import Enum
from pathlib import Path

from ml_pipeline_engine.artifact_store.enums import DataFormat
from ml_pipeline_engine.artifact_store.errors import (
    ArtifactAlreadyExists,
    ArtifactDoesNotExist,
)
from ml_pipeline_engine.artifact_store.serializers import serializer_factory
from ml_pipeline_engine.artifact_store.store.base import SerializedArtifactStore
from ml_pipeline_engine.types import NodeId, NodeResultT, PipelineContextLike


class ArtifactFileAlreadyExists(ArtifactAlreadyExists):
    pass


class ArtifactFileDoesNotExist(ArtifactDoesNotExist):
    pass


def dont_use_for_prod(func: t.Callable):

    @functools.wraps(func)
    async def wrap(*args, **kwargs):
        warnings.warn(f'Функция {func.__name__} предназначена для локального использования')
        return await func(*args, **kwargs)

    return wrap


class FileSystemArtifactStore(SerializedArtifactStore):
    def __init__(self, ctx: PipelineContextLike, artifact_dir: t.Union[Path, str]):
        super().__init__(ctx)

        self.artifact_dir = Path(artifact_dir)

    def _ensure_dir(self) -> Path:
        model_name = self.ctx.model_name.value if isinstance(self.ctx.model_name, Enum) else self.ctx.model_name
        path = Path(self.artifact_dir / model_name / str(self.ctx.pipeline_id))

        if not path.exists():
            os.makedirs(path)

        return path

    def _get_glob(self, node_id: NodeId) -> t.List[Path]:
        return list(Path(self._ensure_dir()).glob(f'{node_id}.*'))

    @dont_use_for_prod
    async def save(self, node_id: NodeId, data: NodeResultT, fmt: DataFormat = DataFormat.PICKLE) -> None:
        if len(self._get_glob(node_id)):
            raise ArtifactFileAlreadyExists(f'Artifact file for {node_id} already exists')

        with open(self._ensure_dir() / f'{node_id}.{fmt.value}', 'wb') as file:
            serializer_factory.from_data_format(fmt).dump(data, file)

    @dont_use_for_prod
    async def load(self, node_id: NodeId) -> NodeResultT:
        glob = self._get_glob(node_id)

        if not len(glob):
            raise ArtifactFileDoesNotExist(f'Artifact file for {node_id} does not exist')

        with open(glob[0], 'rb') as file:
            return serializer_factory.from_extension(glob[0].suffix[1:]).load(file)



import typing as t

import networkx as nx

from ml_pipeline_engine.dag.graph import DiGraph
from ml_pipeline_engine.types import NodeId


def get_connected_subgraph(dag: nx.Graph, source: NodeId, dest: NodeId, is_recurrent: bool = False) -> DiGraph:
    """
    Получить связный подграф между двумя заданными узлами графа
    """

    if len(dag) == 1:
        return t.cast(DiGraph, dag)

    subgraph: DiGraph = dag.subgraph({node_id for path in nx.all_simple_paths(dag, source, dest) for node_id in path})
    subgraph.is_recurrent = is_recurrent

    return subgraph



class BaseBuilderError(Exception):
    pass


class BaseAnnotationError(BaseBuilderError):
    pass


class NonRedefinedGenericTypeError(BaseAnnotationError):
    pass


class UndefinedAnnotation(BaseAnnotationError):
    pass


class UndefinedParamAnnotation(BaseAnnotationError):
    pass


class IncorrectBaseClass(BaseBuilderError):
    pass


class IncorrectTypeClass(BaseBuilderError):
    pass


class IncorrectRecurrentMixinClass(BaseBuilderError):
    pass


class IncorrectParamsRecurrentNode(BaseBuilderError):
    pass



from ml_pipeline_engine.types import NodeBase


class FeatureBase(NodeBase):
    """
    Базовый класс для набора фичей
    """

    node_type = 'feature'

    def extract(self, *args, **kwargs):
        raise NotImplementedError('Method extract() is not implemented')


class FallbackFeatureBase(NodeBase):
    """
    Базовый класс для набора fallback фичей
    """

    node_type = 'feature_fallback'

    def extract(self, *args, **kwargs):
        raise NotImplementedError('Method extract() is not implemented')


class FeatureGenericBase(NodeBase):
    """
    Базовый класс для набора generic фичей
    """

    node_type = 'feature_generic'

    def __init__(self):
        self.value = None

    def extract(self, *args, **kwargs):
        raise NotImplementedError



import typing as t
from dataclasses import dataclass

from ml_pipeline_engine.types import CaseLabel, NodeLike

NodeResultT = t.TypeVar('NodeResultT')


@dataclass(frozen=True)
class InputGenericMark:
    node: NodeLike[t.Any]


@dataclass(frozen=True)
class InputMark:
    node: NodeLike[t.Any]


@dataclass(frozen=True)
class InputOneOfMark:
    nodes: t.List[NodeLike[t.Any]]


def InputOneOf(nodes: t.List[NodeLike[NodeResultT]]) -> t.Type[NodeResultT]:  # noqa
    """
    Принимает список нод, возвращает результат первой успешно выполненной ноды
    """
    return t.cast(t.Any, InputOneOfMark(nodes))


def InputGeneric(node: NodeLike[NodeResultT]) -> t.Type[NodeResultT]:  # noqa
    return t.cast(t.Any, InputGenericMark(node))


def Input(node: NodeLike[NodeResultT]) -> t.Type[NodeResultT]:  # noqa
    return t.cast(t.Any, InputMark(node))


@dataclass(frozen=True)
class GenericInputMark:
    node: NodeLike[t.Any]


def GenericInput(node: NodeLike[NodeResultT]) -> t.Type[NodeResultT]:  # noqa
    return t.cast(t.Any, GenericInputMark(node))


@dataclass(frozen=True)
class SwitchCaseMark:
    switch: NodeLike[t.Any]
    cases: t.List[t.Tuple[str, NodeLike]]
    name: str


def SwitchCase(  # noqa
    switch: NodeLike[t.Any],
    cases: t.List[t.Tuple[CaseLabel, NodeLike[NodeResultT]]],
    name=None,
) -> t.Type[NodeResultT]:
    return t.cast(t.Any, SwitchCaseMark(switch, cases, name))


@dataclass(frozen=True)
class RecurrentSubGraphMark:
    start_node: NodeLike[NodeResultT]
    dest_node: NodeLike[NodeResultT]
    max_iterations: int


def RecurrentSubGraph(  # noqa
    start_node: t.Type[NodeLike[NodeResultT]],
    dest_node: t.Type[NodeLike[NodeResultT]],
    max_iterations: int,
) -> t.Type[NodeResultT]:
    """
    Указание рекуррентного подграфа.

    Задача:
        Исполнить некий подграф `A — B — C` с передачей дополнительных данных в узел `A`
        для изменения поведения всего подграфа и исполнить целевой узел `C` с измененным поведением.

    Инициализация подграфа:
        1. Для инициализации подграфа нужно вызвать self.next_iteration().
        2. Подграф будет исполняться до крайнего случая:
            1. Целевой узел больше не инициализирует self.next_iteration() и достигает необходимых условий исполнения
            2. Количество итераций закончилось и на выход отдается self.get_default()
        3. Узел, с которого начинается подграф, должен иметь в параметрах additional_data для получения данных
           из узла, который инициализировал повторное исполнение подграфа
        4. Инициализация узла возможна только из целевого узла (dest_node). TODO: В будущем мб этот функционал
                                                                                  будет расширен

    Условия:
        1. Целевой подграф должен иметь RecurrentProtocol в базовом классе.
        2. Целевой узел должен определять метод self.get_default()
        3. Целевой узел должен явно задавать use_default=True

    Комбинация с RetryProtocol:
        1. При наличии ошибок системного характера каждый узел может исполниться cls.attempts раз
        2. Классы ошибок можно настроить через cls.exceptions
        3. В случае, если use_default=False и количество попыток исчерпано — будет выкинуто исключение
        4. В случае, если use_default=True — будет вызван метод self.get_default()

    Кейсы:
        1. use_default=True
            1. Если в подграфе в любом узле возникла ошибка, то из узла будет взято дефолтное значение
            2. Если целевой узел достигает условий даже при всех дефолтных значениях предшествующих узлов, то
               этот кейс считается успешным и рекуррентный подграф завершает работу
            3. Если целевой подграф исполняется с ошибкой, то исполняется self.get_default()

        2. use_default=False
            1. Если любой узел подграфа завершается ошибкой, то весь граф завершается ошибкой

    Args:
        start_node: Узел с которого будет начинаться рекуррентный подграф.
        dest_node: Узел которым будет заканчиваться рекуррентный подграф.
        max_iterations: Количество перезапусков подграфа при невыполнении условий.
                        Вызывается с помощью self.next_iteration().
    """
    return t.cast(t.Any, RecurrentSubGraphMark(start_node, dest_node, max_iterations))
